/*
 * Implements DevReg::DeviceRegistry.
 */

#include "libdevreg++.h"
#include <memory>
#include <unistd.h>
#include <set>


using DevReg::Device;
using DevReg::DeviceNode;
using DevReg::ReadException;
using std::set;


void DevReg::DeviceRegistry::cleanup()
{
	vector<Device*>::iterator i = device_list.begin();
	while (i != device_list.end()) {
		delete *i;
		i++;
	}

	device_list.clear();
	deviceid_map.clear();
	deviceidfuzzy_map.clear(); 
	interface_map.clear();
	bus_map.clear();
}

void DevReg::DeviceRegistry::putInMaps(Device *d)
{
	const string &di = d->getDeviceId();
	device_list.push_back(d);
	deviceid_map[di].push_back(d);

	int dsepos  = di.find('/');
	int dsepos2 = di.find('/', dsepos+1);
	int dsepos3 = di.find('/', dsepos2+1);
	string fuzzy = di.substr(0, dsepos) + di.substr(dsepos2, dsepos3-dsepos2);
	deviceidfuzzy_map[fuzzy].push_back(d);

	set<string> interfaces;
	const vector<DeviceNode*> nodes = d->getAllDeviceNodes();
	vector<DeviceNode*>::const_iterator it = nodes.begin();
	while (it != nodes.end()) {
		interfaces.insert((*it)->getInterface());
		it++;
	}

	set<string>::const_iterator is = interfaces.begin();
	while(is != interfaces.end())
		interface_map[*(is++)].push_back(d);
}

void DevReg::DeviceRegistry::crosslink()
{
	vector<Device*> xl;
	map<string, vector<Device*> >::const_iterator it;
	vector<Device*>::iterator i = device_list.begin();
	while (i != device_list.end()) {
		it = deviceid_map.find((*i)->parent_id);
		if (it != deviceid_map.end()) {
			xl = it->second;
			if (xl.size() > 0) {
				xl[0]->children.push_back(*i);
				(*i)->parent = xl[0];
			}
		}

		it = deviceid_map.find((*i)->superdevice_id);
		if (it != deviceid_map.end()) {
			xl = it->second;
			if (xl.size() > 0) {
				xl[0]->subdevices.push_back(*i);
				(*i)->superdevice = xl[0];
			}
		}

		if ((!(*i)->getParent()) && (!(*i)->getSuperDevice())) 
			bus_map[(*i)->getBus()].push_back(*i);

		i++;
	}
}

DevReg::DeviceRegistry::~DeviceRegistry()
{
	cleanup();
}


DevReg::DeviceRegistry::DeviceRegistry(const string &f = "/proc/devreg")
	throw (ReadException) :
	filename(f),
	lastChangecount(-1)
{
	refresh();
}

void DevReg::DeviceRegistry::refresh(int *pfd = 0)
	throw (ReadException)
{
	if (access(filename.c_str(), R_OK))
		throw ReadException("Cannot read devreg file");

	int trynum = 0;
	do { 
		try {
			read(filename, pfd);
		}
		catch (...) {
			cleanup();
			if (trynum >= 10) 
				throw;
			trynum++;
			continue;
		}
	} while(false);
}

void DevReg::DeviceRegistry::read(const string &dir, int *fd)
	throw (ReadException)
{
	int changecount;
	int i = 0;
	if (!exists(dir))
		throw ReadException("DeviceRegistry directory does not exist.");

	changecount = readChangecount(dir + "/changecount", fd);
	if (changecount == lastChangecount)
		return;

	lastChangecount = changecount;
	cleanup();

	while (exists(dir + "/" + i)) {
		putInMaps(new Device(dir + "/" + i + "/"));
		i++;
	}

	if (changecount != readChangecount(dir + "/changecount", fd))
		throw ReadException("DeviceRegistry changed while reading - try again!");

	crosslink();
}

Device *DevReg::DeviceRegistry::getDeviceById(const string &deviceid, 
					      bool onlyExact = false, 
					      bool multiMatch = true) const
{
	vector<Device*> l;
	getDevicesById(deviceid, l, onlyExact);

	if ((l.size() == 0) && (l.size() > 1 && !multiMatch))
		return 0;

	return l[0];
}

bool DevReg::DeviceRegistry::getDevicesById(const string &deviceid, 
					    vector<Device*> &devlist,
					    bool onlyExact = false) const
{
	map<string, vector<Device*> >::const_iterator it = deviceid_map.find(deviceid);
	if (it != deviceid_map.end()) {
		devlist = it->second;
		return devlist.size() > 0;
	}

	devlist.clear();

	if (onlyExact) 
		return false;

	int dsepos  = deviceid.find('/');
	int dsepos2 = deviceid.find('/', dsepos+1);
	int dsepos3 = deviceid.find('/', dsepos2+1);
	string fuzzy = deviceid.substr(0, dsepos) + deviceid.substr(dsepos2, dsepos3-dsepos2);
	string dlocation = deviceid.substr(dsepos, dsepos2);
	string dsernum = deviceid.substr(dsepos3);
	map<string, vector<Device*> >::const_iterator itn = deviceidfuzzy_map.find(fuzzy);
	if (itn == deviceidfuzzy_map.end())
		return false;

	if (dsernum.length() > 1) {
		vector<Device*>::const_iterator vi = itn->second.begin();
		while (vi != itn->second.end()) {
			const string &odevid = (*vi)->getDeviceId();
			int sepos = odevid.find('/');
			int sepos2 = odevid.find('/', sepos+1);
			int sepos3 = odevid.find('/', sepos2+1);
			if (dsernum == odevid.substr(sepos3))
				devlist.push_back(*vi);		
			vi++;
		}
		return false;
	}

	vector<Device*>::const_iterator vi = itn->second.begin();
	while (vi != itn->second.end()) {
		const string &odevid = (*vi)->getDeviceId();
		int sepos  = odevid.find('/');
		int sepos2 = odevid.find('/', sepos+1);
		if (dlocation == odevid.substr(sepos, sepos2))
			devlist.push_back(*vi);		
		vi++;
	}
	
	if (devlist.size() > 0)
		return false;

	devlist = itn->second;
	return false;
}

bool DevReg::DeviceRegistry::getDevicesByInterface(const string &interface, 
						   vector<Device*> &devlist) const
{
	map<string, vector<Device*> >::const_iterator it = interface_map.find(interface);
	
	if (it != interface_map.end()) {
		devlist = it->second;
		return devlist.size() > 0;
	}
	devlist.clear();
	return false;
}

bool DevReg::DeviceRegistry::getDevicesByBus(const string &bus, 
					     vector<Device*> &devlist) const
{
	map<string, vector<Device*> >::const_iterator it;
	it = bus_map.find(bus);
	
	if (it != bus_map.end()) {
		devlist = it->second;
		return devlist.size() > 0;
	}
	devlist.clear();
	return false;
}

bool DevReg::DeviceRegistry::getAllRootDevices(vector<Device*> &devlist) const
{
	devlist.clear();
	int num = 0;
	map<string, vector<Device*> >::const_iterator it = bus_map.begin();
	while (it != bus_map.end()) {
		devlist.insert(devlist.end(), it->second.begin(), it->second.end());
		num = it->second.end() - it->second.begin();
		it++;
	}
	return num > 0;
}


