/*
 * Implements the file reader functions
 */

#include <string>
#include <cstring>
#include <cstdio>
#include <fstream>
#include <fcntl.h>
#include <unistd.h>

#include <libdevreg++.h>

using std::string;
using DevReg::ReadException;

namespace {
	string readStringGeneric(const string &filename, int eof = -1)
		throw (ReadException)
	{
		ifstream f(filename.c_str());
		string s;
		if (f.fail())
			throw ReadException("Cannot open file "+filename);
		if (eof >= 0) {
			getline(f, s);
		}
		else
			getline(f, s);
		if (f.fail())
			throw ReadException("Failure reading file "+filename);
		return s;
	}

	template <class N>
	N readNumber(const string &filename) 
		throw (ReadException) 
	{
		ifstream f(filename.c_str());
		N n;
		if (f.fail())
			throw ReadException("Cannot open file "+filename);
		f >> n;
		if (f.fail())
			throw ReadException("Failure reading file "+filename);
		return n;
	}
}

namespace DevReg {
	string readLine(const string &filename)
		throw (ReadException) 
	{
		ifstream f(filename.c_str());
		string s;
		if (f.fail())
			throw ReadException("Cannot open file "+filename);
		getline(f, s);
		if (f.fail())
			throw ReadException("Failure reading file "+filename);
		return s;
	}

	int readInteger(const string &filename)
		throw (ReadException) 
	{
		return readNumber<int>(filename);
	}

	unsigned long readULong(const string &filename)
		throw (ReadException) 
	{
		return readNumber<unsigned long>(filename);
	}

	int readEnum(const string &filename,
		     const char **values)
		throw (ReadException)
	{
		const char **v = values;
		string s = readStringGeneric(filename);
		int i = 0;
		while (*v) {
			if (!strcmp(*v, s.c_str()))
				return i;
			v++;
			i++;
		}
		throw ReadException("Got unknown enum value "
				    "from "+filename);
	}

	bool exists(const string &filename)
	{
		return access(filename.c_str(), R_OK) ? false : true;
	}

	int readChangecount(const string &filename, 
			    int *pfd = 0)
		throw (ReadException)
	{
		char buf[16];
		int fd;
		bool opened = false;

		if ((!pfd) || ((*pfd) < 0)) {
			fd = open(filename.c_str(), O_RDONLY);
			if (fd < 0)
				throw ReadException("Error opening " +
						    filename);
			opened = true;
		}
		else {
			fd = *pfd;
			if(((int)lseek(fd, 0, SEEK_SET)) == -1)
				throw ReadException("Error lseeking in " +
						    filename);
		}

		// one read is sufficient, implementation guarantees this
		if (read(fd, buf, sizeof(buf)) == -1) {
			if (opened)
				close(fd);
			throw ReadException("Error reading " + filename);
		}
		       
		if (pfd)
			*pfd = fd;
		else
			close(fd);

		return atoi(buf);
	}
}



