/*
 * Simple app to list devices in /proc/devreg
 */	
 
#include <stdio.h> 
#include <getopt.h> 
#include <config.h>
#include <stdlib.h>
#include <stdbool.h>
#include <sys/poll.h>
#include "libdevreg.h"
#include "lsdevreg.h"

static struct option lsdevreg_options[] = 
{
	{"show-subdevices",  0, 0, 's'},
	{"show-devicefiles", 0, 0, 'f'},
	{"exact-match",      0, 0, 'e'},
	{"continuous",       0, 0, 'c'},
	{"devreg-dir",       1, 0, 'r'},
	{"bus",              1, 0, 'b'},
	{"deviceid",         1, 0, 'd'},
	{"interface",        1, 0, 'i'},
	{"help",             0, 0, 'h'},
	{"version",          0, 0, 'v'},
	{0,0,0,0}
};

const char *whitespace = "                                                                                   ";

#define SUBDEVICE_INDENTATION 3
#define CHILDREN_INDENTATION  6

#define CHANGECOUNT "/proc/devreg/changecount"

static void print_help() 
{
	fprintf(stderr, "Usage: lsdevreg [OPTION] [COMMAND]\n"
		"List all devices of the Device Registry.\n"
		"\n"
		"Options:\n"
		"  -s,--show-subdevices       show subdevices\n"
		"  -f,--show-devicefiles      show device files\n"
		"  -e,--exact-match           match only exact device ids\n"
		"  -r,--devreg-dir=FILE       read this device registry directory\n"
		"                             instead of /proc/devreg\n"
		"  -c,--continuous            stay alive and update output if\n"
		"                             the device registry changes\n"
		"Commands:\n"
		"  -b,--bus=NAME              show devices of this bus\n"
		"  -d,--deviceid=NAME         show devices with this deviceid\n"
		"  -i,--interface=NAME        show devices with this interface\n"
		"  -h,--help                  display this help and exit\n"
		"  -v,--version               display version information and exit\n");
}

static void print_version() 
{
	fprintf(stderr, "lsdevreg " LSDEVREG_VERSION "\n"
		"Copyright (C) 2001 Tim O. Jansen \n"
		"\n"
		"See COPYING file for copying conditions (BSD license).\n\n");
}

static void indent(unsigned int indentation) 
{
	ssize_t ileft = (indentation > 80) ? 80 : indentation;

	while (ileft > 0) {
		ssize_t r = fwrite(whitespace, sizeof(char), ileft, stdout);
		if (ileft <= 0)
			return;
		ileft -= r;
	}
}

static void print_device_node(int indentation, 
			      devreg_devicenode_t df) 
{
	indent(indentation);
	printf("*devfile=%s interface=%s\n",
	       devreg_devicenode_get_filename(df),
	       devreg_devicenode_get_interface(df));
}

#define MSUNKNOWN(x) (strlen(x) ? (x) : "(unknown)")

static void print_physical_device(int indentation, 
				  devreg_device_t dev,
				  bool show_files,
				  bool show_subdevices,
				  bool show_children)
{
	indent(indentation);
	printf("name=%s vendor=%s\n",
	       MSUNKNOWN(devreg_device_get_model(dev)),
	       MSUNKNOWN(devreg_device_get_vendor(dev)));
	indent(indentation);
	printf("type=%s bus=%s location=%s\n",
	       MSUNKNOWN(devreg_device_get_type(dev)),
	       MSUNKNOWN(devreg_device_get_bus(dev)),
	       MSUNKNOWN(devreg_device_get_location(dev)));
	indent(indentation);
	printf("deviceid=%s\n", 
	       MSUNKNOWN(devreg_device_get_deviceid(dev)));

	if (show_files) {
		devreg_devicenode_t sdf;
		int i = 0;
		while (i < devreg_device_get_devicenode_num(dev))
			print_device_node(indentation, 
					  devreg_device_get_devicenode(dev, i++));
	}

	if (show_subdevices && (devreg_device_get_subdevice_num(dev) > 0)) {
		int i = 0;

		indent(indentation);
		printf("subdevices:\n");
		while (i < devreg_device_get_subdevice_num(dev)) {
			print_physical_device(indentation+SUBDEVICE_INDENTATION,
					      devreg_device_get_subdevice(dev, i++),
					      show_files,
					      show_subdevices,
					      show_children);
		}
	}

	printf("\n");

	if (show_children)
	{
		int i = 0;
		while (i <  devreg_device_get_child_num(dev)) {
			print_physical_device(indentation+CHILDREN_INDENTATION,
					      devreg_device_get_child(dev, i++),
					      show_files,
					      show_subdevices,
					      show_children);
		}
	}
}

static devreg_t print_device_by_id(const char *filename, 
				   const char *deviceid, 
				   bool show_device_files, 
				   bool show_subdevices, 
				   bool only_exact)
{
	int devnum;
	devreg_t devreg;
	devreg_device_t *devlist;
	int i;
	
	if (!(devreg = devreg_create_registry(filename))) {
		fprintf(stderr, "Cannot read device registry \"%s\".\n", filename);
		return NULL;
	}

	devnum = devreg_get_devices_by_id(devreg, deviceid, NULL, 0, only_exact);
	devlist = alloca(sizeof(devreg_device_t) * devnum);
	if (devnum != devreg_get_devices_by_id(devreg, deviceid, devlist, devnum, only_exact)) {
		fprintf(stderr, "Cannot get expected number of devices.\n");
		return NULL;
	}

	for (i = 0; i < devnum; i++)
		print_physical_device(0, 
				      devlist[i], 
				      show_device_files,
				      show_subdevices,
				      false);

	return devreg;
	
}

static devreg_t print_devices(int (*selectorfunc) (devreg_t, 
						   const char*, 
						   devreg_device_t*,
						   unsigned int),
			      const char *filename, 
			      const char *key,
			      bool show_device_files,
			      bool show_subdevices,
			      bool show_children) 
{
	int devnum;
	devreg_t devreg;
	devreg_device_t *devlist;
	int i;
	
	if (!(devreg = devreg_create_registry(filename))) {
		fprintf(stderr, "Cannot read device registry \"%s\".\n", filename);
		return NULL;
	}
		
	devnum = selectorfunc(devreg, key, NULL, 0);
	devlist = alloca(sizeof(devreg_device_t) * devnum);
	if (devnum != selectorfunc(devreg, key, devlist, devnum)) {
		fprintf(stderr, "Cannot get expected number of devices.\n");
		return NULL;
	}

	for (i = 0; i < devnum; i++)
		print_physical_device(0, 
				      devlist[i], 
				      show_device_files,
				      show_subdevices,
				      show_children);

	return devreg;
}



int main(int argc, char **argv) 
{
	struct pollfd pcc;
	char *filename = "/proc/devreg";
	const char *bus = NULL;
	const char *deviceid = NULL;
	const char *interface = NULL;
	bool show_device_files = false;
	bool show_subdevices = false;
	bool only_exact = false;
	bool continuous = false;
	int last_changecount;
	int fd = -1;
	devreg_t dr;

	int cmds = 0;

	while (1) {
		int c;
		int opt_index = 0;

		c = getopt_long(argc, argv, "-vhr:b:d:i:fsec", lsdevreg_options, &opt_index);

		if (c < 0)
			break;
		switch(c) {
		case 'h':
			print_help();
			return 1;
		case 'v':
			print_version();
			return 1;
		case 'r':
			filename = optarg;
			break;
		case 'f':
			show_device_files = true;
			break;
		case 's':
			show_subdevices = true;
			break;
		case 'e':
			only_exact = true;
			break;
		case 'c':
			continuous = true;
			break;
		case 'b':
			bus = optarg;
			cmds++;
			break;
		case 'd':
			deviceid = optarg;
			cmds++;
			break;
		case 'i':
			interface = optarg;
			cmds++;
			break;
		case '\1':
		case '?':
			fprintf(stderr, "Unexpected argument or option.\n");
			fprintf(stderr, "Use lsdevreg --help for help.\n");
			return 1;
		}

		if (cmds > 1) {
			fprintf(stderr, "lsdevreg: too many commands, try --help for help.\n");
			return 1;
		}
	}

	if (continuous) {
		if (!devreg_read_changecount(CHANGECOUNT, &fd, &last_changecount))
			return 1;
		pcc.fd = fd;
		pcc.events = POLLERR;
	}

	while (1) {
		if (deviceid)
			dr = print_device_by_id(filename, 
						deviceid, 
						show_device_files, 
						show_subdevices, 
						only_exact);
		else if (interface) 
			dr = print_devices(devreg_get_devices_by_interface,
					   filename, 
					   interface, 
					   true,
					   show_subdevices, 
					   false);
		else
			dr =  print_devices(devreg_get_devices_by_bus,
					    filename, 
					    bus, 
					    show_device_files, 
					    show_subdevices, 
					    true);

		if ((!dr) || (!continuous))
			return 1;

		if (!devreg_read_changecount(CHANGECOUNT, &fd, &last_changecount))
			return 1;

		if (poll(&pcc, 1, -1) < 0)
			return 1;

		devreg_refresh(dr);
		printf("\n\nDevice Registry changed:\n");
	}
}
